import datetime
import logging

from schedules_tools.schedule_handlers import ScheduleHandlerBase

log = logging.getLogger(__name__)

try:
    from icalendar import Calendar, Event
    additional_deps_satistifed = True
except ImportError:
    additional_deps_satistifed = False


class ScheduleHandler_ics(ScheduleHandlerBase):
    provide_export = True
    
    default_export_ext = 'ics'
    handle_deps_satisfied = additional_deps_satistifed
    
    _datetime_format = '%Y%m%dT%H%M%SZ'
    _now = None

    def export_schedule(self, out_file, flat=False):
        self._now = datetime.datetime.now()
        
        cal = Calendar()
        cal_params = [
            ('prodid', '-//python-schedules-tools//ICS handler//'),
            ('version', '2.0'),
            ('summary', self.schedule.name),
            ('uid', self.schedule.slug)
        ]
        [cal.add(k, v) for k, v in cal_params]
        
        self.add_tasks_to_calendar(cal, self.schedule.tasks)
        content = cal.to_ical()
        
        with open(out_file, 'wb') as f:
            f.write(content)
        
        return content

    def add_tasks_to_calendar(self, calendar, tasks):
        for task in tasks:
            calendar.add_component(self.task2ical(task))
            self.add_tasks_to_calendar(calendar, task.tasks)

    def task2ical(self, task):
        event = Event()
        event['uid'] = '{}-{}'.format(task.index, task.slug)
        event['summary'] = task.name
        event['dtstart'] = task.dStart.strftime(self._datetime_format)
        event['dtend'] = task.dFinish.strftime(self._datetime_format)
        event['dtstamp'] = self._now.strftime(self._datetime_format)
        event['transp'] = 'TRANSPARENT'  # invisible to free/busy searches      
        event['description'] = task.note
        
        return event




#!/usr/bin/env python

import os
import sys
import logging
from schedules_tools.schedule_handlers.smart_sheet import ScheduleHandler_smartsheet
from multiprocessing import Pool


ss = ScheduleHandler_smartsheet(
    options=dict(smartsheet_token=os.getenv('SMARTSHEET_TOKEN', '')))


def setup_logging(level):
    log_format = '%(name)-10s %(levelname)7s: %(message)s'
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(level)

    formatter = logging.Formatter(log_format)
    sh.setFormatter(formatter)

    # setup root logger
    inst = logging.getLogger('')
    inst.setLevel(level)
    inst.addHandler(sh)


def do_req(worker):
    for i in range(1, 80):
        cont = ss.client.Sheets.list_sheets(include_all=True)
        cont = str(cont)
        if 'Rate limit' in cont:
            print(worker, i)


if __name__ == '__main__':
    setup_logging(logging.WARNING)
    workers = 10

    p = Pool(workers)
    p.map(do_req, range(1, 1 + workers))



import logging
import json
from datetime import datetime

from schedules_tools import jsondate
from schedules_tools.models import Task, Schedule


log = logging.getLogger(__name__)


REPORT_NO_CHANGE = ''
REPORT_ADDED = '_added_'
REPORT_REMOVED = '_removed_'
REPORT_CHANGED = '_changed_'

REPORT_PREFIX_MAP = {
    REPORT_ADDED: '[+]',
    REPORT_REMOVED: '[-]',
    REPORT_CHANGED: '[M]',
    REPORT_NO_CHANGE: 3 * ' ',
}


NAME_SIM_THRESHOLD = 0.8
TASK_SCORE_THRESHOLD = 0.45
NAME_SIM_WEIGHT = 0.5
TASK_POS_WEIGHT = 0.5


def strings_similarity(str1, str2, winkler=True, scaling=0.1):
    """
    Find the Jaro-Winkler distance of 2 strings.
    https://en.wikipedia.org/wiki/Jaro-Winkler_distance

    :param winkler: add winkler adjustment to the Jaro distance
    :param scaling: constant scaling factor for how much the score is adjusted
                    upwards for having common prefixes. Should not exceed 0.25
    """
    if str1 == str2:
        return 1.0

    def num_of_char_matches(s1, len1, s2, len2):
        count = 0
        transpositions = 0  # number of matching chars w/ different sequence order
        limit = int(max(len1, len2) / 2 - 1)

        for i in range(len1):
            start = i - limit
            if start < 0:
                start = 0

            end = i + limit + 1
            if end > len2:
                end = len2

            index = s2.find(s1[i], start, end)

            if index > -1:  # found common char
                count += 1

                if index != i:
                    transpositions += 1

        return count, transpositions

    len1 = len(str1)
    len2 = len(str2)

    num_of_matches, transpositions = num_of_char_matches(str1, len1, str2, len2)

    if num_of_matches == 0:
        return 0.0

    m = float(num_of_matches)
    t = transpositions / 2.0

    dj = (m/float(len1) + m/float(len2) + (m-t)/m) / 3.0

    if winkler:
        length = 0
        # length of common prefix at the start of the string (max = 4)
        max_length = min(
            len1,
            len2,
            4
        )
        while length < max_length and str1[length] == str2[length]:
            length += 1

        return dj + (length * scaling * (1.0 - dj))

    return dj


class ScheduleDiff(object):

    result = []

    hierarchy_attr = 'tasks'
    subtree_hash_attr_name = 'subtree_hash'

    """ List of attributes used to compare 2 tasks. """
    tasks_match_attrs = ['name', 'dStart', 'dFinish', subtree_hash_attr_name]

    def __init__(self, schedule_a, schedule_b, trim_time=False):
        self.schedule_a = schedule_a
        self.schedule_b = schedule_b

        self.trim_time = trim_time

        self.result = self._diff()

    def __str__(self):
        try:
            return unicode(self.result_to_str()).encode('utf-8')
        except NameError:
            return self.result_to_str()

    def _get_subtree(self, item):
        return getattr(item, self.hierarchy_attr)

    def result_to_str(self, items=None, level=0):
        """ Textual representation of the diff. """
        res = ''

        if items is None:
            items = self.result

        schedule = Schedule()

        for item in items:
            subtree = item['subtree']
            state = item['item_state']

            if state in [REPORT_CHANGED, REPORT_ADDED]:
                task = item['right']
            elif state is REPORT_REMOVED:
                task = item['left']
            else:
                task = item['both']

            task_obj = Task.load_from_dict(task, schedule)

            res += '{} {}{}\n'.format(REPORT_PREFIX_MAP[state], level * ' ', str(task_obj))

            if subtree:
                res += self.result_to_str(subtree, level + 2)

        return res

    def _create_report(self,
                       item_state,
                       left=None,
                       right=None,
                       both=None,
                       subtree=[],
                       changed_attrs=[]):
        """
        Returns a dictionary representing a possible change.

            {
                left: Task or None,
                right: Task or None,
                both: used instead of left and right, when the task are equal,
                subtree: List of reports from the child Tasks,
                changed_attr: List of changed attributes,
                item_state: Type of change
            }

        """

        if both:
            report = {
                'both': both.dump_as_dict(recursive=False),
                'subtree': subtree,
                'changed_attrs': changed_attrs,
                'item_state': item_state
            }

        else:
            # No need to keep the whole structure,
            # child tasks will be placed in report['tasks']
            if left is not None:
                left = left.dump_as_dict(recursive=False)

            if right is not None:
                right = right.dump_as_dict(recursive=False)

            report = {
                'left': left,
                'right': right,
                'subtree': subtree,
                'changed_attrs': changed_attrs,
                'item_state': item_state,
            }

        return report

    def _set_subtree_items_state(self, items, state):
        """
        Set the given state recursively on the subtree items
        """

        def create_report(item):
            kwargs = {
                'subtree': self._set_subtree_items_state(self._get_subtree(item), state)
            }

            if state == REPORT_NO_CHANGE:
                kwargs['both'] = item

            elif state == REPORT_ADDED:
                kwargs['right'] = item

            elif state == REPORT_REMOVED:
                kwargs['left'] = item

            return self._create_report(state, **kwargs)

        return [create_report(item) for item in items]

    def get_changed_attrs(self, task_a, task_b):
        """
        Uses attributes defined in `tasks_match_attrs` to compare 2 tasks and
        returns a list of atts that don't match.
        """
        return [attr for attr in self.tasks_match_attrs
                if not self._compare_tasks_attributes(task_a, task_b, attr)]

    def _compare_tasks_attributes(self, task_a, task_b, attr_name):
        """
        Compares tasks attributes.
        Trims time from datetime objects if self.trim_time is set.
        """
        attribute_a = getattr(task_a, attr_name)
        attribute_b = getattr(task_b, attr_name)

        if self.trim_time:
            if isinstance(attribute_a, datetime):
                attribute_a = attribute_a.date()

            if isinstance(attribute_b, datetime):
                attribute_b = attribute_b.date()

        return attribute_a == attribute_b

    def find_best_match(self, t1, possible_matches, start_at_index=0):
        """
        Finds the best match for the given task in the list of possible matches.

        Returns the index of the best match and a dict
            with a state suggestion and list of changed attrs.
        """
        match_index = None
        best_match = {
            'state': REPORT_REMOVED,
            'changes': [],
            'name_score': 0,
            'score': TASK_SCORE_THRESHOLD
        }

        if start_at_index > 0:
            possible_matches = possible_matches[start_at_index:]

        for i, t2 in enumerate(possible_matches, start_at_index):
            res = self.eval_tasks(t1, t2, i, name_threshold=best_match['name_score'])

            if (res['state'] is REPORT_CHANGED
                    and res['score'] > best_match['score']):

                match_index = i
                best_match = res

            if res['state'] is REPORT_NO_CHANGE:
                match_index = i
                best_match = res
                break

        return match_index, best_match

    def _task_position_score(self, index):
        return 1.0 / (2 * (index + 1))

    def _task_score(self, name_score, position_score):
        weight_sum = NAME_SIM_WEIGHT + TASK_POS_WEIGHT
        name_score *= NAME_SIM_WEIGHT
        position_score *= TASK_POS_WEIGHT

        return (name_score + position_score) / weight_sum

    def eval_tasks(self, t1, t2, t2_index, name_threshold=NAME_SIM_THRESHOLD):
        name_score = 0.0
        position_score = 0.0
        changed_attrs = self.get_changed_attrs(t1, t2)

        # different names
        if 'name' in changed_attrs:
            t1_subtree = getattr(t1, self.subtree_hash_attr_name)
            t2_subtree = getattr(t2, self.subtree_hash_attr_name)

            if t1_subtree and t2_subtree:
                if t1_subtree == t2_subtree:
                    state = REPORT_CHANGED
                    position_score = 1.0

                else:
                    name_score = strings_similarity(t1.name, t2.name)

                    if (name_score > name_threshold
                            and len(changed_attrs) < len(self.tasks_match_attrs)):
                        state = REPORT_CHANGED
                        position_score = self._task_position_score(t2_index)
                    else:
                        state = REPORT_REMOVED

            # no subtrees
            else:
                name_score = strings_similarity(t1.name, t2.name, winkler=False)

                if name_score > name_threshold:
                    state = REPORT_CHANGED
                    position_score = self._task_position_score(t2_index)
                else:
                    state = REPORT_REMOVED

        # names are equal
        else:
            name_score = 1.0

            if (not changed_attrs
                or (len(changed_attrs) == 1
                    and self.subtree_hash_attr_name in changed_attrs)):

                state = REPORT_NO_CHANGE
            else:
                state = REPORT_CHANGED
                position_score = 1.0

        return {
            'state': state,
            'changes': changed_attrs,
            'name_score': name_score,
            'position_score': position_score,
            'score': self._task_score(name_score, position_score)
        }

    def _diff(self, tasks_a=None, tasks_b=None):

        if tasks_a is None:
            tasks_a = self.schedule_a.tasks

        if tasks_b is None:
            tasks_b = self.schedule_b.tasks

        res = []
        last_b_index = 0

        # shortcut to create a report for an added task
        def report_task_added(index, recursive=True):
            task = tasks_b[index]
            subtree = self._get_subtree(task)

            if recursive:
                subtree = self._set_subtree_items_state(subtree, REPORT_ADDED)

            return self._create_report(REPORT_ADDED, right=task, subtree=subtree)

        for task in tasks_a:
            match_index, match = self.find_best_match(task, tasks_b, start_at_index=last_b_index)
            report = {}

            if match_index is None:
                subtree = self._set_subtree_items_state(self._get_subtree(task), REPORT_REMOVED)
                report = self._create_report(REPORT_REMOVED, left=task, subtree=subtree)

            else:
                # ALL elements between last_b_index and match_index => ADDED
                res.extend([report_task_added(k) for k in range(last_b_index, match_index)])

                # exact match => NO CHANGE
                if not match['changes']:
                    subtree = self._set_subtree_items_state(self._get_subtree(task), match['state'])
                    report_kwargs = {'both': task, 'subtree': subtree}

                # structural change => CHANGED / NO CHANGE
                elif self.subtree_hash_attr_name in match['changes']:

                    # process child tasks
                    subtree = self._diff(
                        self._get_subtree(task),
                        self._get_subtree(tasks_b[match_index])
                    )

                    if len(match['changes']) > 1:
                        report_kwargs = {
                            'left': task,
                            'right': tasks_b[match_index],
                            'subtree': subtree
                        }

                    else:
                        report_kwargs = {
                            'both': task,
                            'subtree': subtree
                        }

                # no structural changes => CHANGED
                else:
                    subtree = self._set_subtree_items_state(
                        self._get_subtree(tasks_b[match_index]), REPORT_NO_CHANGE)

                    report_kwargs = {
                        'left': task,
                        'right': tasks_b[match_index],
                        'subtree': subtree
                    }

                report = self._create_report(match['state'],
                                             changed_attrs=match['changes'],
                                             **report_kwargs)

                last_b_index = match_index + 1

            res.append(report)

        # remaining tasks => ADDED
        res.extend([report_task_added(k) for k in range(last_b_index, len(tasks_b))])

        return res

    def dump_json(self, **kwargs):

        def _encoder(obj):
            if isinstance(obj, Task):
                return obj.dump_as_dict()
            return jsondate._datetime_encoder(obj)

        kwargs['default'] = _encoder
        return json.dumps(self.result, **kwargs)



import datetime
import logging

from schedules_tools import SchedulesToolsException


log = logging.getLogger(__name__)

try:
    import redis
    redis_available = True
except ImportError:
    log.info('Redis unavailable - will not use exclusive access to storage')
    redis_available = False


class AcquireLockException(SchedulesToolsException):
    pass


class StorageBase(object):
    handle = None
    options = {}
    tmp_root = None
    
    provide_changelog = False
    provide_mtime = False
    
    exclusive_access = False
    exclusive_access_option = 'exclusive_access'
    redis = None    
    
    _shared_lock = None  # shared copy lock
    lock_acquired = None
    lock_timeout = 120
    lock_sleep = 0.5  # seconds
    lock_max_workers = 10  # max number of workers waiting
    

    def __init__(self, handle=None, options=dict()):
        self.handle = handle  # 'handle' is source/target of schedule in general        
        self.options = options
        
        self.exclusive_access = redis_available and options.get(self.exclusive_access_option, 
                                                                self.exclusive_access)

        if self.exclusive_access:
            redis_url = options.get('redis_url', '')
            
            if redis_url:
                self.redis = redis.StrictRedis.from_url(redis_url)
            else:
                self.redis = redis.StrictRedis()

            self._shared_lock = self.redis.lock(
                                name=self.redis_key, 
                                timeout=self.lock_timeout - 10,  # max life time for lock 
                                sleep=self.lock_sleep, 
                                blocking_timeout=self.lock_timeout * self.lock_max_workers
                            )         

    @property
    def redis_key(self):
        return '_'.join(['schedules_tools', self.__class__.__name__])
        

    def acquire_shared_lock(self, msg=''):
        if self.exclusive_access:
            log.debug('Waiting for {} shared lock..'.format(self.__class__.__name__))             
            self.lock_acquired = self._shared_lock.acquire()          

            if not self.lock_acquired:
                raise AcquireLockException(
                    'Unable to acquire lock {}'.format(msg),
                    source=self
                )
            else:
                log.debug('{} shared lock ACQUIRED'.format(self.__class__.__name__))
        
    def release_shared_lock(self):
        if self.lock_acquired:
            self._shared_lock.release()
            log.debug('{} shared lock RELEASED'.format(self.__class__.__name__))  


    def get_local_handle(self, revision=None, datetime=None):
        """
        Get specific version of handle (usually file) based on revision
        or datetime. If they are specified both, revision has precedence.

        Args:
            revision: checkout specific revision
            datetime: checkout regarding to specific date

        Returns:

        """
        raise NotImplementedError

    def clean_local_handle(self):
        raise NotImplementedError

    def push(self):
        raise NotImplementedError    
    
    def get_handle_mtime(self):
        raise NotImplementedError

    def handle_modified_since(self, mtime):
        # Return False only when able to tell
        if isinstance(mtime, datetime.datetime):
            try:
                handle_mtime = self.get_handle_mtime()
            except NotImplementedError:
                return True
            if handle_mtime and handle_mtime <= mtime:
                return False

        return True

    def get_handle_changelog(self):
        raise NotImplementedError


def sync_nfs(local, remote, path=None):
    pass



from schedules_tools import storage_handlers
import os
from datetime import datetime


class StorageHandler_local(storage_handlers.StorageBase):
    provide_mtime = True

    def get_local_handle(self, revision=None, datetime=None):
        return self.handle

    def clean_local_handle(self):
        pass

    def get_handle_mtime(self, path=None):
        mtime_timestamp = os.path.getmtime(self.handle)

        return datetime.fromtimestamp(mtime_timestamp).replace(microsecond=0)



import datetime
import logging
import numbers
import re
import traceback

from schedules_tools.schedule_handlers import ScheduleHandlerBase
from schedules_tools import models, SchedulesToolsException


COLUMN_TASK_NAME = 'name'
COLUMN_DURATION = 'duration'
COLUMN_START = 'start'
COLUMN_FINISH = 'finish'
COLUMN_P_COMPLETE = 'complete'
COLUMN_NOTE = 'note'
COLUMN_FLAGS = 'flags'
COLUMN_LINK = 'link'
COLUMN_PRIORITY = 'priority'

log = logging.getLogger(__name__)


class SmartSheetImportException(SchedulesToolsException):
    pass


class SmartSheetExportException(SchedulesToolsException):
    pass


try:
    import smartsheet
    additional_deps_satistifed = True
except ImportError:
    additional_deps_satistifed = False


class ScheduleHandler_smartsheet(ScheduleHandlerBase):
    provide_export = True
    handle_deps_satisfied = additional_deps_satistifed

    date_format = '%Y-%m-%d'  # 2017-01-20
    datetime_format = '%Y-%m-%dT%H:%M:%S'  # 2017-01-20T08:00:00
    _client_instance = None
    _sheet_instance = None
    _re_number = re.compile('[0-9]+')
    _re_url = re.compile(r'^(https?://.*?smartsheet.com[^?]+)')
    columns_mapping_name = {
        'Task': COLUMN_TASK_NAME,
        'Task Name': COLUMN_TASK_NAME,
        'Duration': COLUMN_DURATION,
        'Start': COLUMN_START,
        'Start Date': COLUMN_START,
        'Due': COLUMN_FINISH,
        'Finish': COLUMN_FINISH,
        'End Date': COLUMN_FINISH,
        '% Complete': COLUMN_P_COMPLETE,
        'Comments': COLUMN_NOTE,
        'Flags': COLUMN_FLAGS,
        'Link': COLUMN_LINK,
        'Priority': COLUMN_PRIORITY,
    }

    columns_mapping_id = {}

    # getter/setter to convert handle to Smartsheet id
    @property
    def handle(self):
        return self._handle

    @handle.setter
    def handle(self, value):
        self._handle = value

        # try to get id if not passed
        if value is not None and not ScheduleHandler_smartsheet.value_is_ss_id(
                                                                    value):
            info_dict = self.get_info_dict(value)

            if 'id' in info_dict:
                self._handle = info_dict['id']

    def get_info_dict(self, value):
        """Return dict with smarsheet id and permalink - get missing one"""

        info_dict = {}

        if ScheduleHandler_smartsheet.value_is_ss_id(value):
            try:
                sheet = self.client.Sheets.get_sheet(value)
                info_dict = dict(id=int(value), permalink=sheet.permalink)
            except smartsheet.exceptions.ApiError as e:
                log.warn(e)

        else:
            match = ScheduleHandler_smartsheet.value_is_ss_permalink(value)
            if match:
                value = match.groups()[0]

                sheets = self.client.Sheets.list_sheets(include_all=True)

                for sheet in sheets.data:
                    if sheet.permalink == value:
                        info_dict = info_dict = dict(id=int(sheet.id),
                                                     permalink=value)
                        break

        return info_dict

    @classmethod
    def value_is_ss_id(cls, value):
        try:
            # looks like a sheet ID (number) - might be string
            int(value)
            return True
        except ValueError:
            pass

    @classmethod
    def value_is_ss_permalink(cls, value):
        return cls._re_url.match(str(value))

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle

        return cls.value_is_ss_permalink(handle) or cls.value_is_ss_id(handle)

    @property
    def client(self):
        if not self._client_instance:
            self._client_instance = smartsheet.Smartsheet(
                self.options['smartsheet_token'])
            self._client_instance.errors_as_exceptions(True)
        return self._client_instance

    @property
    def sheet(self):
        if not self._sheet_instance:
            try:
                # get sheet and turn off rows pagination to get all rows
                self._sheet_instance = self.client.Sheets.get_sheet(
                    self.handle, page_size=None, page=None)

            except smartsheet.exceptions.ApiError as e:
                raise SmartSheetImportException(e.message, source=self.handle)

        return self._sheet_instance

    def get_handle_mtime(self):
        return self.sheet.modified_at

    def get_handle_changelog(self):
        changelog = dict()

        changelog[self.sheet.version] = {
            'date': self.sheet.modified_at,
            'user': None,
            'msg': None,
        }

        return changelog

    def import_schedule(self):
        try:
            self.schedule = models.Schedule()
            self.schedule.name = str(self.sheet.name)
            self.schedule.slug = str(
                self.schedule.unique_id_re.sub('_', self.schedule.name))
            self.schedule.mtime = self.get_handle_mtime()
            self.schedule.changelog = self.get_handle_changelog()
            self.schedule.dStart = datetime.datetime.max
            self.schedule.dFinish = datetime.datetime.min

            parents_stack = []

            # populate columns dict/map
            for column in self.sheet.columns:
                index = self.columns_mapping_name.get(column.title, None)
                if index is None:
                    log.debug('Unknown column %s, skipping.' % column.title)
                    continue
                self.columns_mapping_id[column.id] = index

            for row in self.sheet.rows:
                self._load_task(row, parents_stack)

            if not self.schedule.tasks:
                self.schedule.dStart = None
                self.schedule.dFinish = None
                log.warning('Empty schedule (no valid tasks) %s'
                            % self.get_info_dict(self.handle)['permalink'])

            self.schedule.generate_slugs()

        except (KeyError, IndexError):
            # We can get all exception details via traceback
            # instead of 'except' statement
            msg = traceback.format_exc()
            raise SmartSheetImportException(msg, source=self.handle)

        except smartsheet.exceptions.ApiError as e:
            raise SmartSheetImportException(e.message, source=self.handle)

        return self.schedule

    def _parse_date(self, date_str):
        """
        Method tries to parse date_str into python-native datetime object.

        Native smartsheet columns like 'Start' contains also time, but newly
        added columns via API aren't allowed to store time part.

        Args:
            date_str: Examples of input: 2017-01-20 or 2017-01-20T08:00:00

        Returns:
            datetime instance or None if string couldn't be parsed
        """
        date = None

        for format_str in [self.datetime_format, self.date_format]:
            try:
                date = datetime.datetime.strptime(date_str, format_str)
                break
            except ValueError:
                pass
        # We don't require so precise timestamp, so ignore seconds
        if date:
            date = date.replace(second=0)

        return date

    def _load_task(self, row, parents_stack):
        task = models.Task(self.schedule)
        cells, unknown_cells = self._load_task_cells(row)

        task.index = row.row_number
        # task.slug is generated at the end of importing whole schedule

        # skip empty rows
        if not all([c in cells and cells[c]
                    for c in (COLUMN_TASK_NAME, COLUMN_START, COLUMN_FINISH)]):
            # skip load, task doesn't contain all needed info
            return

        try:
            task.name = unicode(cells[COLUMN_TASK_NAME])
        except NameError:
            task.name = str(cells[COLUMN_TASK_NAME])

        # dates
        task.dStart = self._parse_date(cells[COLUMN_START])
        task.dFinish = self._parse_date(cells[COLUMN_FINISH])

        if not (task.dStart and task.dFinish):
            # Couldn't parse the dates
            return

        self.schedule.dStart = min(self.schedule.dStart, task.dStart)
        self.schedule.dFinish = max(self.schedule.dFinish, task.dFinish)

        if COLUMN_NOTE in cells and cells[COLUMN_NOTE]:
            try:
                task.note = unicode(cells[COLUMN_NOTE])
            except NameError:
                task.note = str(cells[COLUMN_NOTE])

        if COLUMN_PRIORITY in cells:
            task.priority = cells[COLUMN_PRIORITY]

        if COLUMN_DURATION in cells:
            # duration cell contains values like '14d', '~0'
            match = re.findall(self._re_number, cells[COLUMN_DURATION])
            if match:
                task.milestone = int(match[0]) == 0

        if (COLUMN_P_COMPLETE in cells and cells[COLUMN_P_COMPLETE] is not None
                and isinstance(cells[COLUMN_P_COMPLETE], numbers.Number)):
            task.p_complete = round(cells[COLUMN_P_COMPLETE] * 100, 1)

        if COLUMN_FLAGS in cells and cells[COLUMN_FLAGS]:
            # try first to parse workaround format 'Flags: qe, dev'
            task.parse_extended_attr(cells[COLUMN_FLAGS])

            # then try to consider the value as it is 'qe, dev'
            if not task.flags:
                task.parse_extended_attr(cells[COLUMN_FLAGS],
                                         key=models.ATTR_PREFIX_FLAG)

        if COLUMN_LINK in cells and cells[COLUMN_LINK]:
            # try first to parse workaround format 'Link: http://some.url'
            task.parse_extended_attr(cells[COLUMN_LINK])
            # then try to consider the value as it is 'http://some.url'
            if not task.link:
                task.parse_extended_attr(cells[COLUMN_LINK],
                                         key=models.ATTR_PREFIX_LINK)
        # Try to guess column/purpose of unknown cell values
        for cell in unknown_cells:
            task.parse_extended_attr(cell)

        curr_stack_item = {
            'rowid': row.id,
            'task': task
        }

        if not parents_stack:
            # Current task is the topmost (root)
            self.schedule.tasks = [task]
            parents_stack.insert(0, curr_stack_item)

        elif row.parent_id == parents_stack[0]['rowid']:
            # Current task is direct descendant of latest task
            parents_stack[0]['task'].tasks.append(task)
            parents_stack.insert(0, curr_stack_item)

        elif row.parent_id != parents_stack[0]['rowid']:
            # We are currently in the same, or upper, level of tree
            while parents_stack and row.parent_id != parents_stack[0]['rowid']:
                parents_stack.pop(0)

            if parents_stack:
                parents_stack[0]['task'].tasks.append(task)
            else:
                self.schedule.tasks.append(task)

            parents_stack.insert(0, curr_stack_item)
        task.level = len(parents_stack)

    def _load_task_cells(self, row):
        """
        Method tries to combine expected column name together with cell value.
        Cell those don't match are returned back as list of values.

        Args:
            row: smartsheet Row instance

        Returns: tuple (mapped, unknown) cells, resp. unknown values

        """
        mapped_cells = {}
        unknown_values = []
        used_cells = set()

        for cell in row.cells:
            cell_name = self.columns_mapping_id.get(cell.column_id, None)
            if not cell_name:
                if cell.value is not None:
                    unknown_values.append(cell.value)
                continue
            # avoid to override existing values when column name is duplicated
            if cell_name not in used_cells and cell.value:
                mapped_cells[cell_name] = cell.value
                used_cells.add(cell_name)

        return mapped_cells, unknown_values

    def export_schedule(self, output=None):
        # Project sheet from API (Templates.list_public_templates)
        sheet_spec = self.client.models.Sheet({
            'name': self.schedule.name,
            'from_id': 5066554783098756,
        })
        resp = self.client.Home.create_sheet_from_template(sheet_spec)
        self.handle = resp.result.id

        # mark whole week as working days to avoid 'rounding' duration
        # over weekends
        self.client.Sheets.update_sheet(
            self.handle,
            self.client.models.Sheet({
                'project_settings': self.client.models.ProjectSettings({
                    'working_days': [
                        'SUNDAY', 'MONDAY', 'TUESDAY', 'WEDNESDAY', 'THURSDAY',
                        'FRIDAY', 'SATURDAY']
                })
            })
        )

        column_spec_flags = self.client.models.Column({
            'title': 'Flags',
            'type': 'TEXT_NUMBER',
            'index': 4,
        })
        column_spec_link = self.client.models.Column({
            'title': 'Link',
            'type': 'TEXT_NUMBER',
            'index': 4,
        })
        resp = self.client.Sheets.add_columns(
            self.handle,
            [column_spec_flags, column_spec_link]
        )

        if resp.message != 'SUCCESS':
            msg = 'Adding column failed: {}'.format(resp)
            raise SmartSheetExportException(msg, source=self.handle)

        for task in self.schedule.tasks:
            self.export_task(task, parent_id=None)

        # sheet ID
        return self.handle

    def export_task(self, task, parent_id=None):
        row = self.client.models.Row()
        row.parent_id = parent_id
        row.to_bottom = True

        if task.name:
            row.cells.append({
                'column_id': self.sheet.columns[0].id,
                'value': task.name
            })
        if task.dStart:
            row.cells.append({
                'column_id': self.sheet.columns[2].id,
                'value': task.dStart.isoformat()
            })
        if task.milestone:
            row.cells.append({
                'column_id': self.sheet.columns[1].id,
                'value': '0'
            })
        elif task.dFinish:
            # finish date can be set only as (start, duration) tuple,
            # put direct value for Finish column is not allowed.
            # Even if start and finish date are the same, duration equals 1.
            duration = (task.dFinish - task.dStart).days + 1
            row.cells.append({
                'column_id': self.sheet.columns[1].id,  # finish #3
                'value': '{}d'.format(duration)
            })
        if task.flags:
            row.cells.append({
                'column_id': self.sheet.columns[4].id,
                'value': ', '.join(task.flags)
            })
        if task.link:
            row.cells.append({
                'column_id': self.sheet.columns[5].id,
                'value': task.link
            })
        if task.p_complete:
            row.cells.append({
                'column_id': self.sheet.columns[8].id,
                'value': task.p_complete / 100.0
            })
        if task.note:
            row.cells.append({
                'column_id': self.sheet.columns[10].id,
                'value': task.note
            })
        resp = self.client.Sheets.add_rows(self.handle, [row])

        if resp.message != 'SUCCESS':
            raise SmartSheetExportException(resp.result.message,
                                            source=self.handle)
        len_result = len(resp.result)
        if 1 != len_result:
            msg = ('Just one row was expected to be added, instead '
                   'of {}'.format(len_result))
            raise SmartSheetExportException(msg, source=self.handle)

        row_id = resp.result[0].id

        for nested_task in task.tasks:
            self.export_task(nested_task, parent_id=row_id)

        return row_id



#!/usr/bin/env python

import os
import sys

from setuptools import setup, find_packages

from scripts.setup_utils import read_version, write_version, get_rpm_version


project_name = "schedules-tools"
project_url = "https://github.com/RedHat-Eng-PGM/python-schedules-tools"
project_author = "Red Hat, Inc."
project_author_email = "pp-dev-list@redhat.com"
project_description = "Schedules tools to handle/convert various schedule formats"
package_name = project_name


# VERSION - write version only when building from git
package_version = ['5.30.0', 1, 'git']  # default
save_version_dirs = ["schedules_tools"]

if os.path.isdir(".git"):
    # we're building from a git repo -> store version tuple to __init__.py
    if package_version[2] == "git":
        force = True
        package_version[0], package_version[1], package_version[2] = get_rpm_version()

    for i in save_version_dirs:
        file_name = os.path.join(i, "version.py")
        write_version(file_name, package_version)
else:  # see if there is already saved version
    for i in save_version_dirs:
        file_name = os.path.join(i, "version.py")
        version = read_version(file_name)
        if version:
            package_version = list(version)
            break


py_version = sys.version[:1]

setup(
    name=package_name,
    url=project_url,
    version='.'.join([str(v) for v in package_version[0:2]]),
    license='GPL',
    author=project_author,
    author_email=project_author_email,
    description=project_description,
    packages=find_packages(exclude=('scripts',)),
    include_package_data=True,
    entry_points={
        # Make them all start with schedule- and end with python version
        #   so we can generate symlinks in /usr/bin/ post install
        'console_scripts': [
            'schedule-convert%s=schedules_tools.bin.schedule_convert:main' % py_version,
            'schedule-diff%s=schedules_tools.bin.schedule_diff:main' % py_version,
        ]
    },
    install_requires=[
        'icalendar',
        'lxml'
    ],
    extras_require={
        "Rally":  ["pyral"],
        "Smartsheet": ["smartsheet-python-sdk<3"],
    },
    tests_require=[
        'mock',
        'pytest',
        'redis',
        'testtools',
    ],
)



import logging
import os

from schedules_tools import discovery
from schedules_tools import SchedulesToolsException
from schedules_tools.discovery import search_paths
from schedules_tools.models import Schedule
from schedules_tools.storage_handlers import AcquireLockException


log = logging.getLogger(__name__)


SORT_FIELDS = {
    'name': 'name',
    'date_start': 'dStart',
    'date_finish': 'dFinish',
    'source': None,
}


class ScheduleFormatNotSupported(Exception):
    pass


class StorageFormatNotSupported(Exception):
    pass


class HandleWithoutExport(Exception):
    pass


class HandlerMissingDeps(Exception):
    pass


class ScheduleConverter(object):
    """
    Abstraction class to work with handles/schedules
    no matter the exact handler/schedule type.
    """
    schedule = None
    storage_handler = None
    local_handle = None

    def __init__(self, schedule=None):
        self.schedule = schedule

    # TODO: take a look if _get_handler* methods can't be shared for both
    # schedule/storage
    @staticmethod
    def _get_schedule_handler_for_handle(handle):
        for module in discovery.schedule_handlers.values():
            if (module['class'].handle_deps_satisfied and
                    module['class'].is_valid_source(handle)):
                return module

        msg = "Can't find schedule handler for handle: {}".format(handle)
        raise ScheduleFormatNotSupported(msg)

    @staticmethod
    def _get_schedule_handler_for_format(fmt):
        if fmt not in discovery.schedule_handlers.keys():
            msg = "Can't find schedule handler for format: {}".format(fmt)
            raise ScheduleFormatNotSupported(msg)

        return discovery.schedule_handlers[fmt]

    @staticmethod
    def _get_storage_handler_for_format(fmt):
        if fmt not in discovery.storage_handlers.keys():
            msg = "Can't find storage handler for format: {}".format(fmt)
            raise StorageFormatNotSupported(msg)

        return discovery.storage_handlers[fmt]

    @classmethod
    def _get_schedule_handler_struct(cls, handle=None, fmt=None):
        if fmt:
            handler_struct = cls._get_schedule_handler_for_format(fmt)
        else:
            handler_struct = cls._get_schedule_handler_for_handle(handle)

        return handler_struct

    @classmethod
    def _get_schedule_handler_cls(cls, *args, **kwargs):
        return cls._get_schedule_handler_struct(*args, **kwargs)['class']

    @classmethod
    def _get_storage_handler_cls(cls, *args, **kwargs):
        return cls._get_storage_handler_for_format(*args, **kwargs)['class']

    def _init_storage_handler(self,
                              handle,
                              storage_src_format,
                              options=dict()):
        """Init storage handler if it's necessary and isn't already prepared"""
        if storage_src_format and not self.storage_handler:
            storage_handler_cls = self._get_storage_handler_cls(
                                                        storage_src_format)
            storage_handler = storage_handler_cls(handle=handle,
                                                  options=options)
            self.storage_handler = storage_handler

        return self.storage_handler

    def _get_local_handle_from_storage(self,
                                       handle,
                                       storage_src_format,
                                       options):
        """
        Init storage and get local handle from storage
        if specified otherwise return original handle
        """
        if not self.local_handle:
            self.local_handle = handle

            self._init_storage_handler(handle, storage_src_format, options)

            if self.storage_handler:
                self.local_handle = self.storage_handler.get_local_handle()

        return self.local_handle

    def cleanup_local_handle(self):
        if self.storage_handler:
            self.storage_handler.clean_local_handle()

        self.local_handle = None

    # Following methods call their counterparts on handlers
    def handle_modified_since(self,
                              handle,
                              mtime,
                              schedule_src_format=None,
                              storage_src_format=None,
                              cleanup=True,
                              options=dict()
                              ):
        """
        Return boolean (call schedule_handler specific method)
        to be able to bypass processing
        """
        errors = []
        handle_modified = True

        self._init_storage_handler(handle, storage_src_format, options)

        # use storage if possible, otherwise schedule
        if self.storage_handler and self.storage_handler.provide_mtime:
            handle_modified = self.storage_handler.handle_modified_since(mtime)
        else:
            local_handle = self._get_local_handle_from_storage(
                handle, storage_src_format, options)

            schedule_handler_cls = self._get_schedule_handler_cls(
                                            handle=local_handle,
                                            fmt=schedule_src_format
                                            )

            if not schedule_handler_cls.handle_deps_satisfied:
                msg = (
                    'Schedule handler {} has unsatisfied dependencies, '
                    'can\'t get modified time.'.format(schedule_handler_cls))
                raise HandlerMissingDeps(msg)

            try:
                handle_modified = schedule_handler_cls(
                    handle=local_handle, options=options
                    ).handle_modified_since(mtime)
            except SchedulesToolsException as e:
                error_item = e.__class__.__name__, str(e).split('\n'), e.source
                errors.append(error_item)

            if cleanup:
                self.cleanup_local_handle()

        return handle_modified, errors

    def import_schedule(self,
                        handle,
                        schedule_src_format=None,
                        storage_src_format=None,
                        cleanup=True,
                        options=dict()
                        ):

        # convert to local handle if needed
        local_handle = self._get_local_handle_from_storage(handle,
                                                           storage_src_format,
                                                           options
                                                           )
        schedule_handler_cls = self._get_schedule_handler_cls(
                                        handle=local_handle,
                                        fmt=schedule_src_format
                                        )
        if not schedule_handler_cls.handle_deps_satisfied:
            msg = ('Schedule handler {} has unsatisfied dependencies, '
                   'can\'t import schedule.'.format(schedule_handler_cls))
            raise HandlerMissingDeps(msg)

        schedule_handler = schedule_handler_cls(
                                        handle=local_handle,
                                        options=options
                                        )

        # imports changelog and mtime - if implemented
        schedule = Schedule()
        try:
            schedule = schedule_handler.import_schedule()

            # if storage defined and provides changelog/mtime
            # - use storage handler to overwrite it
            if self.storage_handler:
                if self.storage_handler.provide_changelog:
                    schedule.changelog = self.storage_handler.get_handle_changelog()

                if self.storage_handler.provide_mtime:
                    schedule.mtime = self.storage_handler.get_handle_mtime()

        except SchedulesToolsException as e:
            error_item = e.__class__.__name__, str(e).split('\n'), e.source
            schedule.errors_import.append(error_item)

        self.schedule = schedule

        if cleanup:
            self.cleanup_local_handle()

        return self.schedule

    def export_schedule(self,
                        output,
                        target_format,
                        update_filename=False,
                        options=dict()):
        schedule_handler_cls = self._get_schedule_handler_cls(
                                                        fmt=target_format)

        if not schedule_handler_cls.provide_export:
            raise HandleWithoutExport(
                'Schedule handler for {} doesn\'t provide export.'
                .format(target_format))

        if not schedule_handler_cls.handle_deps_satisfied:
            msg = ('Schedule handler {} has unsatisfied dependencies, '
                   'can\'t export schedule.'.format(schedule_handler_cls))
            raise HandlerMissingDeps(msg)

        schedule_handler = schedule_handler_cls(schedule=self.schedule,
                                                options=options)

        if update_filename and schedule_handler.default_export_ext:
            # change/add export extension according to handler
            output = '.'.join([os.path.splitext(output)[0],
                               schedule_handler.default_export_ext])

        return schedule_handler.export_schedule(output)


def convert(args):
    opt_args = vars(args)

    # set timezone - it's desirable that all calculations are made in same TZ
    os.environ['TZ'] = opt_args['tz']

    for path in opt_args.pop('handlers_path'):
        search_paths.append(path)

    converter = ScheduleConverter()

    try:
        converter.import_schedule(
            handle=args.source,
            schedule_src_format=args.source_format,
            storage_src_format=args.source_storage_format,
            options=opt_args)

        if converter.schedule.errors_import:
            for err in converter.schedule.errors_import:
                msg = '{} Handle: {}\n{}'.format(
                        err[0], err[2], '\n'.join(err[1]))
                log.error(msg)
            return False

    except AcquireLockException as e:
        log.error(e)
        return

    check_tasks = dict()
    for task_name in args.check_taskname:
        check_tasks[task_name] = False

    for task_name in args.check_taskname_startswith:
        check_tasks[task_name] = True

    if args.check_taskname or args.check_taskname_startswith:
        missing_tasks = converter.schedule.check_for_taskname(check_tasks)
        if missing_tasks:
            log.info('Missing tasks: {}'.format(list(missing_tasks)))

    if args.flat or args.milestones:
        converter.schedule.make_flat()

    if args.milestones:
        converter.schedule.filter_milestones()

    if args.sort:
        converter.schedule.sort_tasks(args.sort)

    flag_show = args.flag_show.split(',')
    if flag_show == ['']:
        flag_show = []

    flag_hide = [f for f in args.flag_hide.split(',') if f]
    if flag_hide == ['']:
        flag_hide = []

    converter.schedule.filter_flags(flag_show, flag_hide)

    # do we have target name defined?
    update_filename = False
    if not args.target:
        args.target = args.source
        update_filename = True

    converter.export_schedule(args.target,
                              args.target_format,
                              update_filename=update_filename,
                              options=opt_args)


def get_handlers_args_parser(add_help=False):
    """Return parent parser for schedules tools scripts with handler arguments"""
    import argparse

    parser = argparse.ArgumentParser(add_help=add_help)

    parser.add_argument('--log-level',
                        help='ERROR (default) | WARN | INFO | DEBUG',
                        default='ERROR')

    parser.add_argument(
        '--handlers-path',
        help='Add python-dot-notation path to discover handlers (needs to '
             'be python module), can be called several times '
             '(conflicting names will be overriden - the last '
             'implementation will be used)',
        action='append',
        default=[],
    )

    parser.add_argument('-f', '--force',
                        help='Force target overwrite',
                        default=False,
                        action='store_true')

    parser.add_argument('--slug', metavar='SCHEDULE_SLUG',
                        help='Override schedule slug (e.g. rhel)')
    parser.add_argument('--use-tji-file',
                        help='Use TJI file when exporting into TJP',
                        default=False,
                        action='store_true')
    parser.add_argument(
        '--tjp-keep-tjx',
        help='Keep tjx intermediate file when importing from TJP',
        default=False,
        action='store_true')

    parser.add_argument('--rally-iter', help='Rally iteration to import',
                        default='')

    parser.add_argument('--check-taskname',
                        help='Check existence given task name as exact match '
                             '(can be used multiple times)',
                        action='append', metavar='TASKNAME', default=[])
    parser.add_argument(
        '--check-taskname-startswith',
        help='Check existence given task by matching beginning '
             'name (can be used multiple times)',
        action='append', metavar='TASKNAME_STARTSWITH',
        default=[])

    parser.add_argument('--source-storage-format',
                        choices=discovery.storage_handlers.keys(),
                        metavar='SRC_STORAGE_FORMAT',
                        help='Source storage format to use')

    parser.add_argument('--cvs-repo-name',
                        help='Name of CVS repository to checkout')
    parser.add_argument('--cvs-root',
                        help='Root of CVS repository')
    parser.add_argument('--cvs-checkout-path',
                        help='Path to shared working copy of CVS repository')
    parser.add_argument(
        '--cvs-exclusive-access',
        help='Restrict to run just one CVS command at the same time',
        action='store_true')
    parser.add_argument(
        '--cvs-lock-redis-uri',
        help='Redis URI that is required by --cvs-exclusive-access, '
             'default: localhost:6379/0',
        default='localhost:6379/0')

    parser.add_argument('--smartsheet-token',
                        help='Access token for using SmartSheet API')

    parser.add_argument('--date-format',
                        help='Date format used for export where applicable')

    parser.add_argument('--html-title',
                        help='HTML export page title')
    parser.add_argument('--html-table-header',
                        help='HTML export table header')

    parser.add_argument('--flat',
                        help='Make output schedule flat',
                        default=False,
                        action='store_true')

    parser.add_argument('--milestones',
                        help='Filter only milestones (implies --flat)',
                        default=False,
                        action='store_true')

    parser.add_argument(
        '--flag-show',
        help='Filter schedule - show tasks with any of these flags',
        default='',
    )
    parser.add_argument(
        '--flag-hide',
        help='Filter schedule - hide tasks with any of these flags, '
             'hide has preferrence over show',
        default='',
    )

    parser.add_argument('--source-format',
                        metavar='SRC_FORMAT',
                        help='Source format to enforce')

    parser.add_argument('--target-format',
                        metavar='TARGET_FORMAT',
                        help='Target format to convert',
                        default='html')
    parser.add_argument('--tz',
                        help='Timezone used for schedule conversions (default "America/New_York")',
                        default='America/New_York')

    def sorting_field(value):
        try:
            field = SORT_FIELDS[value]
        except KeyError:
            raise argparse.ArgumentTypeError('"%s" is not a valid sorting value' % value)

        return field

    parser.add_argument('--sort',
                        help='Sort by: %s' % ', '.join(SORT_FIELDS.keys()),
                        type=sorting_field)

    return parser
